package com.vernaio.example;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.AnomalyDetectionData;
import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.AnomalyDetectionInference;
import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.AnomalyDetectionSetup;
import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.AnomalyDetectionTraining;
import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.dto.DtoInferenceConfig;
import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.dto.DtoTargetConfig;
import com.vernaio.aivis.engine.v2.anomaly.detection.sdk.dto.DtoTrainingConfig;
import com.vernaio.aivis.engine.v2.base.sdk.EngineLogger;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFilesFilter;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoInferenceTimeseriesDataFilter;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoInterval;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoTimeseriesFilesReaderConfig;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoInferenceDataSpecification;

public class Main {
	
	private static final int MIN_LOGGING_LEVEL = 3;
	
	private static final int MAX_THREADS = 0; // max threads 0 means the number of threads will be chosen automatically
	
	private static final String TRAINING_DATA_FILE = "train_ad.csv";
	
	private static final String INFERENCE_DATA_FILE = "eval_ad.csv";
	
	private static final String TARGET_SIGNAL_ID = "L-1";
	
	private static final long INFERENCE_START_TIME = 1465632000000L;
	
	private static final long INFERENCE_END_TIME = 1466786640000L;
	
	public static void main(final String[] args) throws Exception {
		
		System.out.println("Setup ...");
		// parsing args as folder config
		final FolderConfig folderConfig = FolderConfig.fromArgs(args);
		// create and register logger
		AnomalyDetectionSetup.registerLogger(new EngineLogger() {
			
			@Override
			public void log(int level, String thread, String module, String message) {
				if (level <= MIN_LOGGING_LEVEL) {
					System.out.println(String.format("\t... %s", message));
				}
			}
		});
		// init thread count
		AnomalyDetectionSetup.initThreadCount(MAX_THREADS);
		System.out.println("\t... done!");
		
		/*
		 * TRAINING
		 */
		
		System.out.println("Loading training data ...");
		
		final String[] trainingFiles = { TRAINING_DATA_FILE };
		final DtoFilesFilter trainingFilesFilter = new DtoFilesFilter().withIncludeFiles(trainingFiles);
		
		final DtoTimeseriesFilesReaderConfig trainingFilesReaderConfig =
			new DtoTimeseriesFilesReaderConfig(folderConfig.getInputFolder()).withFilesFilter(trainingFilesFilter);
		
		final AnomalyDetectionData trainingData = AnomalyDetectionData.create();
		trainingData.readFiles(trainingFilesReaderConfig);
		
		System.out.println("Creating training ...");
		// build training config - for more options here, see documentation
		final DtoTrainingConfig trainingConfig = new DtoTrainingConfig(new DtoTargetConfig(TARGET_SIGNAL_ID));
		
		// create training and train the model
		final AnomalyDetectionTraining training = AnomalyDetectionTraining.create(trainingData, trainingConfig);
		System.out.println("\t... done!");
		
		System.out.println("Retrieving and saving training files ...");
		// model is needed for inference and also contains the inference data specification
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "model.json"), training.getModel());
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "training-report.json"), training.getReport());
		System.out.println("\t... done!");
		
		/*
		 * INFERENCE
		 */
		
		// saving memory, we don't need training data anymore
		trainingData.destroy();
		System.out.println("Loading inference data ...");
		
		final String[] inferenceFiles = { INFERENCE_DATA_FILE };
		final DtoFilesFilter inferenceFilesFilter = new DtoFilesFilter().withIncludeFiles(inferenceFiles);
		
		final IDtoInferenceDataSpecification dataSpec = training.getModel().getInferenceDataSpecification();
		final DtoInferenceTimeseriesDataFilter inferenceDataFilter =
			new DtoInferenceTimeseriesDataFilter(
				new IDtoInferenceDataSpecification[] { dataSpec },
				new DtoInterval[] { new DtoInterval().withStartTime(INFERENCE_START_TIME).withEndTime(INFERENCE_END_TIME) }
			).withAdditionalSignals(new String[] { TARGET_SIGNAL_ID });
		
		final DtoTimeseriesFilesReaderConfig inferenceFilesReaderConfig =
			new DtoTimeseriesFilesReaderConfig(folderConfig.getInputFolder()).withFilesFilter(inferenceFilesFilter).withDataFilter(inferenceDataFilter);
		
		final AnomalyDetectionData inferenceData = AnomalyDetectionData.create();
		inferenceData.readFiles(inferenceFilesReaderConfig);
		
		System.out.println("Creating inference ...");
		// build inference config - for more options here, see documentation
		// skipOnInsufficientData: if inference data does not meet the inference data specification for a timestamp an inference is demanded ("insufficient data"),
		// no exception is thrown but it's simply "skipped"
		final DtoInferenceConfig inferenceConfig = new DtoInferenceConfig(true);
		// note that we use here try-with-resource notation to destroy inference handle, inference.destroy() works equivalently
		try (final AnomalyDetectionInference inference = AnomalyDetectionInference.createByTraining(training, inferenceConfig)) {
			
			training.destroy();
			System.out.println("\t... done!");
			
			System.out.println("Calculating scores ...");
			// here, we choose exactly the timestamp that are in the inference data file
			final List<Long> inferenceTimestamps = new ArrayList<Long>();
			for (long time = INFERENCE_START_TIME; time < INFERENCE_END_TIME; time += 60000L) {
				inferenceTimestamps.add(time);
			}
			final List<? extends IDtoFloatDataPoint> scores = inference.infer(inferenceData, inferenceTimestamps);
			System.out.println("\t... done!");
			
			/*
			 * comment in you want to calculate next normal point
			 * System.out.println("Calculating scores with next normal point ...");
			 * // here, we choose exactly the timestamp that are in the inference data file
			 * final List<Long> inferenceTimestampsNextNormal = new ArrayList<Long>();
			 * for (long time = 1465632000000L; time < 1466786640000L; time += 600000L) {
			 * inferenceTimestampsNextNormal.add(time);
			 * }
			 * final List<? extends IDtoFloatDataPointWithNextNormal> scoresWithNextNormal =
			 * inference.inferWithNextNormal(inferenceData, inferenceTimestampsNextNormal, 0.5);
			 * System.out.println("\t... done!");
			 */
			
			System.out.println("Saving inference files ...");
			Utils.writeScores(new File(folderConfig.getOutputFolder(), "scores.json"), scores);
			// comment in if scoresWithNextNormal were computed to save as json
			// Utils.writeScoresWithNextNormal(new File(folderConfig.getOutputFolder(), "scores-with-next-normal.json"), scoresWithNextNormal);
			inferenceData.destroy();
			System.out.println("\t... done!");
			
		}
	}
}
